package worldwinddemo;

// XML
import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
// IO
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
// WW
import gov.nasa.worldwind.*;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.layers.*;

/**
 * @author Patrick Murris - patmurris.blogspot.com
 * @version 0.2 - 20070524 - Added support for wms servers
 * history  0.1 - 20070523 - Very first try at a XML images layer loader
 * Reads a XML WW layer definition and create the corresponding TiledImagelayer
 * ----------------------------------------------------------------------------
 * Issues : doesnt support layer hierarchy (path) to a particular layer.
 * Layers load even when still far away - have to figure out number of empty levels.
 * Cache path has 'Earth' hardcoded.
 * ----------------------------------------------------------------------------
 */

public class WWXMLLayer  extends TiledImageLayer {
	
    public WWXMLLayer(String fileName, String layerName)
    {
        super(makeLevels(readXML(new File(fileName)), layerName));
        this.setName(layerName);
        this.setUseTransparentTextures(true);        
    }

    // Read an XML document from a file
    private static Document readXML(File file)
    {
        if (file == null)
        {
            String message = WorldWind.retrieveErrMsg("nullValue.FileIsNull");
            WorldWind.logger().log(java.util.logging.Level.FINE, message);
            throw new IllegalArgumentException(message);
        }

        try
        {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilderFactory.setNamespaceAware(false);
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            Document doc = docBuilder.parse(file);
            return doc;
        }
        // TODO: use proper error message strings - not GeoRSS
        catch (ParserConfigurationException e)
        {
            String message = WorldWind.retrieveErrMsg("GeoRSS.ParserConfigurationException");
            WorldWind.logger().log(java.util.logging.Level.FINE, message, e);
            throw new WWRuntimeException(message, e);
        }
        catch (IOException e)
        {
            String message = WorldWind.retrieveErrMsg("GeoRSS.IOExceptionParsing") + file.getPath();
            WorldWind.logger().log(java.util.logging.Level.FINE, message, e);
            throw new WWRuntimeException(message, e);
        }
        catch (SAXException e)
        {
            String message = WorldWind.retrieveErrMsg("GeoRSS.IOExceptionParsing") + file.getPath();
            WorldWind.logger().log(java.util.logging.Level.FINE, message, e);
            throw new WWRuntimeException(message, e);
        }
    }
    
    // Create LevelSet from xml doc for the specified layerName
    private static LevelSet makeLevels(Document doc, String layerName)
    {
    	// Get QuadtileSet node
        Node node = findQuadtileSetByName(doc, layerName);
        if(node == null)
        	return null; // TODO: throw exception

        // Extract useful info
        Node child;
        // Sector
        Sector sector = Sector.FULL_SPHERE;
        child = findChildByName(node, "BoundingBox");
        if(child != null) {
        	Node n;
          	n = findChildByName(child, "North");
        	double north = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "South");
        	double south = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "West");
        	double west = Double.parseDouble(findChildByName(n, "Value").getTextContent());
          	n = findChildByName(child, "East");
        	double east = Double.parseDouble(findChildByName(n, "Value").getTextContent());
        	sector = new Sector(Angle.fromDegrees(south), Angle.fromDegrees(north), Angle.fromDegrees(west), Angle.fromDegrees(east));
        }
        // Images and server access
        Node imageTileService, wmsAccessor;
        double levelZeroTileSizeDegree;
        int numLevels;
        int tileSize = 512;
        String service, datasetName, formatSuffix;
        String wmsServiceName = "", wmsVersion = "", wmsImageFormat = "", wmsTransparency = "", wmsStyle = "";
        child = findChildByName(node, "ImageAccessor");
        if(child != null) {
        	Node n;
          	n = findChildByName(child, "LevelZeroTileSizeDegrees");
          	levelZeroTileSizeDegree = Double.parseDouble(n.getTextContent());
          	n = findChildByName(child, "NumberLevels");
          	numLevels = Integer.parseInt(n.getTextContent());
          	n = findChildByName(child, "TextureSizePixels");
          	tileSize = Integer.parseInt(n.getTextContent());
          	n = findChildByName(child, "ImageFileExtension");
          	formatSuffix = "." + n.getTextContent();
          	imageTileService = findChildByName(child, "ImageTileService");
          	wmsAccessor = findChildByName(child, "WMSAccessor");
          	if(imageTileService != null) {
          		// WW image tile service
          		n = imageTileService;
          		Node n2;
              	n2 = findChildByName(n, "ServerUrl");
              	service = n2.getTextContent();
              	n2 = findChildByName(n, "DataSetName");
              	datasetName = n2.getTextContent();
          	} else if(wmsAccessor != null) {
          		// WMS server
          		n = wmsAccessor;
          		Node n2;
              	n2 = findChildByName(n, "ServerGetMapUrl");
              	service = n2.getTextContent();
              	n2 = findChildByName(n, "WMSLayerName");
              	datasetName = n2.getTextContent();
              	n2 = findChildByName(n, "ServiceName");
              	if(n2 != null) wmsServiceName = n2.getTextContent();
              	n2 = findChildByName(n, "Version");
              	if(n2 != null) wmsVersion = n2.getTextContent();
              	n2 = findChildByName(n, "ImageFormat");
              	if(n2 != null) wmsImageFormat = n2.getTextContent();
              	n2 = findChildByName(n, "UseTransparency");
              	if(n2 != null) wmsTransparency = n2.getTextContent();          		
              	n2 = findChildByName(n, "WMSLayerStyle");
              	if(n2 != null) wmsStyle = n2.getTextContent();          		
          	} else {
          		// No ImageTileService or WMSAccessor node
          		// TODO: throw exception
          		return null;
          	}
        } else {
        	// No Image accessor node
        	// TODO: throw exception
        	return null;
        }
        
        // Compose level param
        AVList params = new AVListImpl();
        params.setValue(Level.TILE_WIDTH, tileSize);
        params.setValue(Level.TILE_HEIGHT, tileSize);
        params.setValue(Level.CACHE_NAME, "Earth/" + layerName); // TODO: get cache root from somewhere
        params.setValue(Level.SERVICE, service);
        params.setValue(Level.DATASET_NAME, datasetName);
        params.setValue(Level.FORMAT_SUFFIX, formatSuffix);
        params.setValue(Level.NUM_LEVELS, numLevels);
        params.setValue(Level.NUM_EMPTY_LEVELS, 0); // TODO: compute empty levels from levelZeroTileSize
        params.setValue(Level.LEVEL_ZERO_TILE_DELTA, new LatLon(Angle.fromDegrees(levelZeroTileSizeDegree), Angle.fromDegrees(levelZeroTileSizeDegree)));
        params.setValue(AVKey.SECTOR, sector);
        // WMS params
        if(wmsAccessor != null) {
        	// TODO: use static Level string constants
            params.setValue("WMS_SERVICE_NAME", wmsServiceName);
            params.setValue("WMS_VERSION", wmsVersion);
            params.setValue("WMS_IMAGE_FORMAT", wmsImageFormat);
            params.setValue("WMS_TRANSPARENCY", wmsTransparency);
            params.setValue("WMS_STYLE", wmsStyle);
            params.setValue(Level.TILE_URL_BUILDER, new URLBuilderWMS());
        }

        // Return LevelSet
        return new LevelSet(params);
    }

    private static class URLBuilderWMS implements Level.TileURLBuilder
    {
        public URL getURL(Tile tile) throws MalformedURLException
        {
            StringBuffer sb = new StringBuffer(tile.getLevel().getService());
            if (sb.lastIndexOf("?") != sb.length() - 1)
                sb.append("?");
            sb.append("version=");
            sb.append(tile.getLevel().getParams().getValue("WMS_VERSION"));
            sb.append("&request=GetMap");
            sb.append("&layers=");
            sb.append(tile.getLevel().getDataset());
            sb.append("&srs=EPSG:4326");
            sb.append("&width=");
            sb.append(tile.getLevel().getTileWidth());
            sb.append("&height=");
            sb.append(tile.getLevel().getTileHeight());

            Sector s = tile.getSector();
            sb.append("&bbox=");
            sb.append(s.getMinLongitude().getDegrees());
            sb.append(",");
            sb.append(s.getMinLatitude().getDegrees());
            sb.append(",");
            sb.append(s.getMaxLongitude().getDegrees());
            sb.append(",");
            sb.append(s.getMaxLatitude().getDegrees());

            sb.append("&format=");
            sb.append(tile.getLevel().getParams().getValue("WMS_IMAGE_FORMAT"));
            sb.append("&styles=");
            sb.append(tile.getLevel().getParams().getValue("WMS_STYLE"));
            // TODO: sort out what parameters are required and in which case...
            //sb.append("&service=");
            //sb.append(tile.getLevel().getParams().getValue("WMS_SERVICE_NAME"));
            //sb.append("&transparent=");
            //sb.append(tile.getLevel().getParams().getValue("WMS_TRANSPARENCY"));

            //System.out.println(sb.toString());
            
            return new java.net.URL(sb.toString());
        }
    }

    @Override
    public String toString()
    {
        return "WWXMLLayer";
    }

    // DOM tools
 
    // Find a QuadtileSet node by its name - exact match case sens.
    private static Node findQuadtileSetByName(Document doc, String layerName) {
        NodeList nodes = doc.getElementsByTagName("QuadTileSet");
        if(nodes.getLength() < 1)
        	return null;
        // Find quadtileset node for layerName
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            Node name = findChildByName(node, "Name"); 
            if(name != null && name.getTextContent().equals(layerName))
            	return node;
        }
        return null;
    }
 
    // Find a child node by its name - not case sensitive
    private static Node findChildByName(Node parent, String name)
    {
        NodeList children = parent.getChildNodes();
        if (children == null || children.getLength() < 1)
            return null;

        for (int i = 0; i < children.getLength(); i++)
        {
            String n = children.item(i).getNodeName();
            if (n != null && n.toLowerCase().equals(name.toLowerCase()))
                return children.item(i);
        }

        return null;
    }

}
